require "defines"
require "migration"

require "mod_def"

if DF_Under_Debug then
  require "_tools.debug_tool"
  DF_Print=1
--  DF_DrawBuildMarkers=1
--DF_MakeItems=1
end  

local SLIPSTREAM_VERSION = "1.1.0"
local sschests = {}
local energy_per_action = 3000

-- 3k - 16 rounds 
-- 4k - 12 
-- 5k   10 

local next = next
local north = defines.direction.north
local east = defines.direction.east
local south = defines.direction.south
local west = defines.direction.west

game.on_init(function()
	onLoad()
end)

game.on_load(function()
	onLoad()
  if DF_MakeItems == 1 then
    startingItems()
   end
end)

game.on_save(function()
	global.slipstream.sschests = sschests
end)


game.on_event(defines.events.on_built_entity, function(event) OnBuiltEntity(event) end)
game.on_event(defines.events.on_built_entity, function(event) OnBuiltEntity(event) end)
game.on_event(defines.events.on_entity_died, function(event) OnEntityDied(event) end)
game.on_event(defines.events.on_preplayer_mined_item, function(event) OnPrePlayerMinedItem(event) end)
game.on_event(defines.events.on_tick, function(event) OnTick(event) end)
game.on_event(defines.events.on_player_rotated_entity, function(event) OnRotated(event) end)
game.on_event(defines.events.on_robot_pre_mined, function(event) OnPrePlayerMinedItem(event) end)
--



function IsSlipstreamChest(_Entity)
  return  string.find(_Entity.name, "^slipstream%-chest%-[%a%-]+$")
end

function OnTick(_Event)
  if string.match(game.tick,'^%d+(0)$') then
    local needClean=false
    for i,v in ipairs(sschests) do
      if v.entity.valid then

        if not v.catchArea then
          slipstreamConfigureArea(v)
        end

        if v.beltMask == 0 then
          slipstreamSearchBelts(v)
        end
        
        if v.beltMask ~= 0 then
          slipstreamTick(v)
        end
        
      else
        --entityRemoved()
        needClean=true
      end
    end
    if needClean then
      cleanEntitys()            
    end
  end
end

function OnBuiltEntity(_Event)
  if IsSlipstreamChest(_Event.created_entity) then
    slipstreamBuilt(_Event.created_entity)
  else
    if _Event.created_entity.type=="transport-belt" then
      slipstreamOnBuiltBelt(_Event.created_entity)
    end
  end
end

function OnEntityDied(_Event)
  if IsSlipstreamChest(_Event.entity) then
    slipstreamRemove(_Event.entity)
  else
    if _Event.entity.type=="transport-belt" then
      slipstreamOnRemoveBelt(_Event.entity)
    end
  end
end

function OnPrePlayerMinedItem(_Event)
  if IsSlipstreamChest(_Event.entity) then
    slipstreamRemove(_Event.entity)
  else
    if _Event.entity.type=="transport-belt" then
      slipstreamOnRemoveBelt(_Event.entity)
    end
  end
end

function OnRotated(_Event)
  if _Event.entity.type=="transport-belt" then
    slipstreamOnRotateBelt(_Event.entity)
  end
end

function onLoad()
	if not global.slipstream then
		global.slipstream = {}
		global.slipstream.sschests = {}
		global.slipstream.version = SLIPSTREAM_VERSION
	end
	
	if global.slipstream.version ~= SLIPSTREAM_VERSION then
		slipstreamMigration(global.slipstream.version, SLIPSTREAM_VERSION)
	end
	
	sschests = global.slipstream.sschests
  
  for i,v in ipairs(sschests) do -- to migrate
    if v.entity.valid then
      if not v.catchArea then
        slipstreamConfigureArea(v)
      end
    end
  end
	
end

function cleanEntitys()
	for i,v in ipairs(sschests) do
		if not v.entity.valid then
			v.battery.destroy()
			table.remove(sschests,i)
		end
	end
end

function slipstreamRemove(_Entity)
  for i,v in ipairs(sschests) do
    if not v.entity.valid or v.entity==_Entity then
      v.battery.destroy()
      table.remove(sschests,i)
      return
    end
  end
end

function slipstreamBeltUpdate(_Entity,op)
  for i,v in ipairs(sschests) do
    if not v.entity.valid then
      v.battery.destroy()
      table.remove(sschests,i)
    else
      if posInArea(_Entity.position,v.catchArea) then
--        debugLog("ssc belt catch "..tostring(op).." "..tostring(v.beltMask).." nrc "..#v.neighborBelts)
        v.beltMask=0
        if op == 0 then
          table.insert(v.neighborBelts,_Entity)
        elseif op==1 then
          v.neighborBelts={} -- full rescan
        elseif op == 2 then
          local fake=1
           v.neighborBelts={} -- full rescan
        end
--        slipstreamSearchBelts(v)
--        debugLog("ssc after belt catch "..tostring(op).." "..tostring(v.beltMask).." nrc "..#v.neighborBelts)
      end  
    end
  end
end

function slipstreamOnBuiltBelt(_Entity)
  slipstreamBeltUpdate(_Entity,0)
end

function slipstreamOnRemoveBelt(_Entity)
  slipstreamBeltUpdate(_Entity,1)
end

function slipstreamOnRotateBelt(_Entity)
  slipstreamBeltUpdate(_Entity,2)
end


-- Construct the data table for each chest as it's built.
function slipstreamBuilt(entity)
	local chest_name = entity.name
	local _Surface    = entity.surface
	
	local new_chest = {}
--	new_chest.output_config = {}
	
	-- Record position of chest and search area.
	local pos = {x = entity.position.x, y = entity.position.y}
	local search = {{pos.x - 0.5,pos.y - 0.5},{pos.x + 0.5, pos.y + 0.5}}
	
	entity.destroy() -- Remove initial chest.
	
	-- Creates the accumulator entity that serves as the chest's power supply.
	_Surface.create_entity{name="slipstream-chest-battery", position = pos, force = game.forces.player}
	
	-- Recreate the chest on top of the accumulator.
	_Surface.create_entity{name = chest_name, position = pos, force = game.forces.player}
	
	-- Find the entity references for the battery/chest.
	local t = _Surface.find_entities_filtered{area = search, name = "slipstream-chest-battery"}
	new_chest.battery = t[1]
	new_chest.battery.minable = false
	new_chest.battery.destructible = false
	local t = _Surface.find_entities_filtered{area = search, name = chest_name}
	new_chest.entity = t[1]
	new_chest.inventory = t[1].get_inventory(defines.inventory.chest)
	

--	new_chest.output_config.target = 1
--	new_chest.output_config.limit   =1

  slipstreamConfigureArea(new_chest)	
	table.insert(sschests, new_chest)
end

function slipstreamConfigureArea(_Chest)
  if not _Chest.catchArea then
    local _Surface    = _Chest.entity.surface
    local pos = {x = _Chest.entity.position.x, y = _Chest.entity.position.y}
    local searchBelt = {{pos.x - 1,pos.y - 1},{pos.x + 1, pos.y + 1}}
    
--    _Surface.create_entity{name="m2k-dbg-small-red", position = searchBelt[1], force = game.forces.player}
--    _Surface.create_entity{name="m2k-dbg-small-blue", position = searchBelt[2], force = game.forces.player}
    _Chest.catchArea=searchBelt
    _Chest.itemCounts={}
    _Chest.neighborBelts={}
    _Chest.inBelts={}
    _Chest.outBelts={}
    _Chest.beltMask=0
    _Chest.beltState=0
    _Chest.collectIndex=1
    _Chest.spawnIndex=1
  end 
end

function slipstreamSearchBelts(_Chest)
  if #_Chest.neighborBelts == 0 then
    local _Surface    = _Chest.entity.surface
    _Chest.neighborBelts=_Surface.find_entities_filtered{area = _Chest.catchArea, type = "transport-belt"}
  end
  
  if #_Chest.neighborBelts == 0 then
    return false
  end
  
  _Chest.beltMask=0
  _Chest.inBelts={}
  _Chest.outBelts={}

  local pos = {x = _Chest.entity.position.x, y = _Chest.entity.position.y}


  local beltscan_coords = { -- Points to search for transport belts.
    [north] = {pos.x    , pos.y -1},
    [east] =  {pos.x + 1, pos.y},
    [south] = {pos.x    , pos.y + 1},
    [west] =  {pos.x - 1, pos.y}
     }
  local directions = {north, east, south, west} -- For the for loop.
  
  for belt_idx,belt in ipairs(_Chest.neighborBelts) do
    for direction_idx,cdir in ipairs(directions) do
      if belt.position.x == beltscan_coords[cdir][1] and belt.position.y == beltscan_coords[cdir][2] then
        if cdir == belt.direction then -- output belt
          table.insert(_Chest.outBelts,belt)
          _Chest.beltMask=bit32.bor(_Chest.beltMask,bit32.lshift (1, cdir/2))
        else
          table.insert(_Chest.inBelts,belt)
          _Chest.beltMask=bit32.bor(_Chest.beltMask,bit32.lshift (1, cdir/2+4))
        end
         table.remove(directions,direction_idx)
      end  
    end
  end
  
  if _Chest.beltMask ~= _Chest.beltState then
    _Chest.itemCounts={}
    _Chest.collectIndex=1
    _Chest.spawnIndex=1
  end
end

function posInArea(_Pos,_Area) -- normalize area ?!
  local state=_Pos.x >= _Area[1][1] and _Pos.x <= _Area[2][1] and
         _Pos.y >= _Area[1][2] and _Pos.y <= _Area[2][2]
  
  return state
end

function slipstreamTick(_Chest)
  local maxIters=math.floor(_Chest.battery.energy / energy_per_action) --2*#_Chest.outBelts + #_Chest.inBelts*2 +5
--  debugLog("iters "..tostring(maxIters))
  local last_energy=_Chest.battery.energy
  while _Chest.battery.energy  > energy_per_action do
    local ci=_Chest.collectIndex
    local moved = false
    local tries = 0

    if maxIters== 0 then
      return
    end
    -- collect items
    while not moved and tries < #_Chest.inBelts*2 do
      local li=_Chest.collectIndex 
      local srcBelt = _Chest.inBelts[math.floor   ((li+1)/2)]
      if not srcBelt.valid then
         _Chest.beltMask=0
         _Chest.inBelts={}
         _Chest.outBelts={}
         _Chest.neighborBelts={}
        return
      end

      local lane=srcBelt.get_transport_line((li % 2) +1)
      tries=tries+1
      
      for res_name,res_count in pairs(lane.get_contents()) do
        local r_stack={name = res_name, count =1}
        if _Chest.inventory.can_insert(r_stack) then
          _Chest.inventory.insert(r_stack)
          lane.remove_item (r_stack)
          _Chest.battery.energy = _Chest.battery.energy - energy_per_action
          moved=true
          break
        end
      end
      
      _Chest.collectIndex=_Chest.collectIndex + 1 
      if _Chest.collectIndex >= (2*#_Chest.inBelts+1) then
        _Chest.collectIndex=1
      end   
    end 

    if _Chest.battery.energy  < energy_per_action then
      return -- no power
    end 
    --spawn item
    moved = false
    tries = 0
    for res_name,res_count in pairs(_Chest.inventory.get_contents()) do
      
      local r_stack={name = res_name, count =1}

      while not moved and tries < #_Chest.outBelts*2 do
        local li=_Chest.spawnIndex 
        local targetBelt = _Chest.outBelts[math.floor   ((li+1)/2)]
        if not targetBelt.valid then
           _Chest.beltMask=0
           _Chest.inBelts={}
           _Chest.outBelts={}
           _Chest.neighborBelts={}
          return
        end
        local targetLane=  targetBelt.get_transport_line(li % 2 +1)
        tries = tries + 1
        if targetLane.can_insert_at_back() then
          _Chest.inventory.remove(r_stack)
          targetLane.insert_at_back(r_stack)
          _Chest.battery.energy = _Chest.battery.energy - energy_per_action
          moved=true
        end
        _Chest.spawnIndex=_Chest.spawnIndex + 1 
        if _Chest.spawnIndex >= (2*#_Chest.outBelts+1) then
          _Chest.spawnIndex=1
        end   
      end
        
      if moved then
        break
      end
    end
    
    maxIters=maxIters-1 -- guard
    if last_energy==_Chest.battery.energy then 
      return -- collect and spawn loops does nothing
    end
    last_energy=_Chest.battery.energy
  end
end

function startingItems()
  game.player.character.insert{name="smartsplitter", count=50}
  game.player.character.insert{name="slipstream-chest-basic", count=50}

  game.player.character.insert{name="iron-plate", count=1000}
  game.player.character.insert{name="copper-plate", count=1000}

  game.player.character.insert{name="small-electric-pole", count=32}
  game.player.character.insert{name="basic-inserter", count=64}
  game.player.character.insert{name="solar-panel", count=54}
  game.player.character.insert{name="basic-transport-belt", count=128}
  game.player.character.insert{name="steam-engine", count=16}
  game.player.character.insert{name="boiler", count=32}
  game.player.character.insert{name="lab", count=8}
  game.player.character.insert{name="pipe", count=64}
  game.player.character.insert{name="basic-mining-drill", count=32}
  game.player.character.insert{name="basic-transport-belt-to-ground", count=32}
  game.player.character.insert{name="pipe-to-ground", count=32}
  game.player.character.insert{name="basic-splitter", count=32}
  game.player.character.insert{name="coal", count=128}
  game.player.character.insert{name="raw-wood", count=128}
  game.player.character.insert{name="car", count=1}
  game.player.character.insert{name="chemical-plant", count=8}
  game.player.character.insert{name="assembling-machine-3", count=8}
end

